-- Alternative version of command.e
-- allows you to execute this several times without unexecuting.
-- command.e
--
-- Ben Lin 5/98
--
--	(Command Pattern)
--
--	Commands are the only objects allowed to change the blackboard's contents
--	via an interface defined in the blackboard class. Die is a good example
--	of a command - it is defined for all blackboards.
--	Other commands might look like: add_queen(6, 3) (add a queen to the square
--	(6, 3) on a chessboard)
--
-- 7/6/98
--		Redesigned so it uses plain_execute, plain_unexecute
--	I wanted to keep the names execute, and unexecute because I
--	believe the default behaviour should be to (un)execute
--	all forced moves as soon as any step is (un)done.
--
-- 9/6/98
--		Minor changes (I only found out about elseif recently!). Tidied
--	code, fixed a bug with useless commands.
--
-- 10/6/98
--		Fixed an unexecute bug - it wasn't setting all the executed flags
--	to false. Made no_contradiction_unexecute to make it look nicer
--
-- 14/6/98
--		Allowed multiple executions of a command. What if
-- I want to increase an integer on the blackboard several times?
-- So I changed executed, useless to integers.

deferred class COMMAND

feature
	execute is
	-- execute command, get forced moves (via observers) and execute them too
	local
		i : INTEGER
	do
		if bb.alive then
			-- If dead, nothing happens. Not even the history gets updated
			executed := executed + 1
			if finished then
				-- useless command still must be recorded so we can unexecute it
				useless := useless + 1
				bb.command_history.add_last(Current)
			elseif can_execute then
				bb.command_history.add_last(Current)
				plain_execute
				!!forced_move.make
				append_forced
				from i := forced_move.lower
				until i > forced_move.upper or bb.dead
				-- optimization: if dead, don't bother
				-- executing the rest of the forced moves
				loop
					forced_move.item(i).execute
					i := i + 1
				end
			else
				bb.suicide
				contradiction := true
				bb.command_history.add_last(Current)
			end
		end

	-- cannot ensure executed > 0 - commands have no effect when dead
	end


	unexecute is
		-- undo up to and including this command
	require
		executed > 0

	local
		c : COMMAND
	do
		if contradiction then
			bb.rebirth
			contradiction := false
			bb.command_history.remove_last
		else
			from c := bb.command_history.last
			until c = Current
			loop
				c.no_contradiction_unexecute
				-- could put `unexecute' here, but it's
				-- more efficient this way (contradictions can
				-- only happen once at the end, and we also know
				-- `c' is last on the history)

				c := bb.command_history.last
			end
			no_contradiction_unexecute
		end

	ensure
		not contradiction
	end

	plain_execute is
		-- Guts of `execute'. Modifies blackboard here.
	deferred end

	plain_unexecute is
		-- Counterpart of `plain_execute'.
	deferred end

	append_forced is
	require
		forced_move /= Void
	deferred end
		-- call observers, and append forced moves to the list

	can_execute : BOOLEAN is
		-- true if it is possible to execute the command
	deferred end

	finished : BOOLEAN is
		-- true if the command has been executed, or if another
		-- command with the same effect has been executed in the past.
	deferred end

	display is
		-- display information about command
		-- useful for debugging
	do
		io.put_string("command%N")
	end

	forced_move : COMMAND_LIST
		-- list of forced moves

	put_forced_move(new_forced_move : COMMAND_LIST) is
		-- set the list of forced moves to `new_forced_move'
		-- need this for an efficient MULTICOMMAND `append_forced'
	do
		forced_move := new_forced_move
	end

	bb : BLACKBOARD
		-- blackboard this command acts on

	contradiction : BOOLEAN
		-- true if this command contradicts with a previous command

	useless : INTEGER
		-- increases each time this command has no effect

	executed : INTEGER
		-- increases each time this command is executed

feature {OBSERVER}
	forced_die is
		-- Adds a DIE command to the head of the `forced_move' list
	local
		c : DIE
	do
		!!c.make(bb)
		forced_move.add_first(c)
	end

feature {COMMAND}

	no_contradiction_unexecute is
		-- called by 'unexecute': unexecutes assuming
		-- it's last in the history, doesn't check for contradictions
	require
		not contradiction
		bb.command_history.last = Current
		executed > 0
	do
		executed := executed - 1
		if useless > 0 then
			useless := useless - 1
		else
			plain_unexecute
			-- `c' is last, so this is allowed
		end
		bb.command_history.remove_last
	ensure
		not contradiction
	end

feature {NONE}

	notify(observer_list : OBSERVER_LIST) is
		-- for use in the concrete versions
	local
		i : INTEGER
	do
		if not observer_list.empty then
			from i := observer_list.lower
			until i > observer_list.upper
			loop
				observer_list.item(i).update(Current)
				i := i + 1
			end
		end
	end

end
