-- Logic Problems Blackboard III
--
-- Blackboards I, II are available on request - they're not very good,
-- as they use a much smaller board. For harder logic problems,
-- they have to do actual searching to find a solution.
--
-- Also, I introduced the LOGIC_GRID generic class,
-- and a class called TICK_GRID, which is almost LOGIC_GRID[INTEGER]
-- Ben Lin 26/5/98
--
-- See some examples for more detail, such as method.e

class LP_BOARD
inherit BLACKBOARD
	redefine display
creation make
feature
	board : TICK_GRID

	size : INTEGER
		-- the number of different values an attribute can take
		-- each subboard is a `size' x `size' grid

	sub_count : INTEGER
		-- the number of attributes

	tick_count : INTEGER
		-- counts the number of ticks placed so far

	tick_target : INTEGER
		-- the number of ticks needed for a solution

	tick_observer : LOGIC_GRID[OBSERVER_LIST]
	cross_observer : like tick_observer
		-- local observers
		-- every square on the grid has two observer lists
		-- one list gets notified if the square is ticked, and the
		-- other does if it's crossed

	all_cross_observer : like all_tick_observer
	all_tick_observer : OBSERVER_LIST
		-- global observers
		-- these get notified for any tick or cross.
		-- strictly speaking, not needed - can put one on each
		-- tick_observer, but it's more efficient this way.

	make(subboard_count, size_of_subboard : INTEGER) is
	local
		i : INTEGER
		j : INTEGER
		x : INTEGER
		y : INTEGER
		one_tick : DISTINCT
		elim : ELIMINATION
		prop_tick : PROPAGATE_TICK
		prop_cross : PROPAGATE_CROSS
		obs_list : OBSERVER_LIST
	do
		-- every creation clause of a blackboard needs this:
		!!command_history.make

		size := size_of_subboard
		sub_count := subboard_count

		-- calculate the number of ticks required for a solution:
		-- number of subboards = number of ways of picking 2 attributes
		-- from `sub_count' attributes (= `sub_count' choose 2)
		-- and each subboard has `size' ticks.
		tick_target := (size * sub_count * (sub_count - 1) ) // 2
		!!board.make(sub_count, size)
		!!tick_observer.make(sub_count, size)
		!!cross_observer.make(sub_count, size)
		
		-- These global observers are usually needed in every logic problem
		-- (of course, these lists can be modified)
		!!all_tick_observer.make
		!!all_cross_observer.make

		!!one_tick.make
		!!prop_tick.make
		!!elim.make
		!!prop_cross.make
		all_tick_observer.add_last(one_tick)
		all_tick_observer.add_last(prop_tick)
		all_cross_observer.add_last(elim)
		all_cross_observer.add_last(prop_cross)

		-- Initialize all the local observer lists (they start off empty)
		from j := 2
		until j > sub_count
		loop
			from i := 1
			until i = j
			loop
				from x := 1
				until x > size
				loop
					from y := 1
					until y > size
					loop
						!!obs_list.make
						tick_observer.put(obs_list, i, j, x, y)

						!!obs_list.make
						cross_observer.put(obs_list, i, j, x, y)
						y := y + 1
					end
					x := x + 1
				end

				i := i + 1
			end
			j := j + 1
		end
	end

	finished : BOOLEAN is
	do
		Result := tick_count = tick_target and alive
	end

	get_open : like command_history is
	local
		tick : ADD_TICK
		j, x, y : INTEGER
	do
		!!Result.make
		if alive then
			from j := 2
			until j > sub_count
			loop
				from x := 1
				until x > size
				loop
					from y := 1
					until y > size
					loop
						if board.item(1, j, x, y) = 0 then
							!!tick.make(Current, 1, j, x, y)
							Result.add_last(tick)
						end
						y := y + 1
					end
					x := x + 1
				end
				j := j + 1
			end
		end

	end

	display is
	local
		i, j, x, y : integer
	do
		from i := 1
		until i = sub_count
		loop
			from x := 1
			until x > size
			loop
				from j := i + 1
				until j > sub_count
				loop
					from y := 1
					until y > size
					loop
						inspect board.item(i, j, x, y)
						when 1 then
							io.put_character('O')
						when 0 then
							io.put_character('.')
						when -1 then
							io.put_character('x')
						else
							io.put_character('?')
						end
						y := y + 1
					end
					io.put_character(' ')
					j := j + 1
				end

				io.put_character('%N')
				x := x + 1
			end
			io.put_character('%N')
			i := i + 1
			
		end -- loop

		if not alive then
			io.put_string("CONSTRAINT VIOLATED%N")
		end -- if
	end

feature {COMMAND}

	-- Interface to `COMMAND' class, this is what commands
	-- call to modify the current state of the solution.

	add_tick(i, j : INTEGER; x, y : INTEGER) is
	require
		board.item(i, j, x, y) = 0
	do
		board.tick(i, j, x, y)
		tick_count := tick_count + 1
	end

	add_cross(i, j : INTEGER; x, y : INTEGER) is
	require
		board.item(i, j, x, y) = 0
	do
		board.cross(i, j, x, y)
	end

	remove_tick(i, j : INTEGER; x, y : INTEGER) is
	require
		board.item(i, j, x, y) = 1
	do
		board.blank(i, j, x, y)
		tick_count := tick_count - 1
	end

	remove_cross(i, j : INTEGER; x, y : INTEGER) is
	require
		board.item(i, j, x, y) = -1
	do
		board.blank(i, j, x, y)
	end

end 

